# -*- coding: utf-8 -*-
"""
Created on Tue Feb 19 14:31:08 2019

@author: Iain Cruickshank
"""

import numpy as np, multiprocessing as mp, networkx as nx
from scipy.sparse import csr_matrix, lil_matrix, eye, issparse, diags
from sklearn.neighbors import kneighbors_graph
from sklearn.preprocessing import normalize
from sknetwork.clustering import Louvain, modularity
from sklearn.utils.validation import check_symmetric
from itertools import starmap
    

'''
Unimodal Graph Leanring Techniques
''' 
    
class consensus_graph:
    def __init__(self, metric='euclidean', tau=0, input_graph=None, 
                 multi_process = False, n_processes=4, k=0.4, symmetrize=True):
        
        self.metric=metric
        self.tau = tau
        self.multi_process = multi_process
        self.n_processes = n_processes
        self.input_graph = input_graph
        self.k=k
        self.symmetrize = symmetrize
            
            
    def learn_graph(self, data_matrix=None):
        
        if self.input_graph is None:
            A= data_matrix
            N = A.shape[0]
            self.input_graph = kneighbors_graph(A, metric = self.metric, mode='connectivity', n_neighbors=int(np.floor(N*self.k)))
            if self.symmetrize:
                self.input_graph = self.input_graph.minimum(self.input_graph.T)
        else:
            N = self.input_graph.shape[0]
            
        neighbors = {vertex:self.input_graph.getrow(vertex).nonzero()[1] for vertex in range(N)}
            
        if self.multi_process:
            with mp.Pool(processes=self.n_processes) as pool:
                graphs = pool.starmap(self._score_edges, 
                                               [(vertex, neighbors, N) for vertex in range(N)])
        else:
            graphs = [self._score_edges(vertex, neighbors, N) for vertex in range(N)]
                        
        final_graph = normalize(np.sum(graphs, axis=0), norm='l1', axis=1, copy=False)
        final_graph.data[final_graph.data < self.tau] = 0
        final_graph.eliminate_zeros()
        return final_graph
    
    
    def _score_edges(self, vertex, neighbors, N):
        C = lil_matrix((N,N))
        for p in range(N):
            for q in range(p+1,N):
                if (p in neighbors[vertex]) and (q in neighbors[vertex]):
                    C[p,q] +=1 
                    C[q,p] +=1
                        
        return csr_matrix(C)
    
    
    
class weighted_consensus_graph:
    def __init__(self, metric='euclidean', scoring='bray_curtis', epsilon=0, 
                 input_graph=None, multi_process = False, n_processes=4, k=0.4, symmetrize=True):
        self.metric = metric
        self.scoring = scoring
        self.epsilon = epsilon
        self.multi_process = multi_process
        self.n_processes = n_processes
        self.input_graph = input_graph
        self.k = k
        self.symmetrize = True
        
        
    def learn_graph(self, data_matrix=None):
        
        if self.input_graph is None:
            A= data_matrix
            N = A.shape[0]
            kNN = kneighbors_graph(A, metric = self.metric, mode='distance', n_neighbors=int(np.floor(N*self.k)))
            if self.symmetrize:
                kNN = kNN.minimum(kNN.T)
            self.input_graph = dist_to_sim().convert(kNN)
        else:
            N = self.input_graph.shape[0]
            
        neighbors = {vertex:self.input_graph.getrow(vertex).nonzero()[1] for vertex in range(N)}
        vectors = {vertex:self.input_graph[vertex] for vertex in range(N)}
        scorer = scoring()
        scored_graph = np.sum([scorer.score(vertex, neighbors, vectors, N, self.scoring) for vertex in range(N)])
        
        if self.multi_process:
            with mp.Pool(processes=self.n_processes) as pool:
                graphs = pool.starmap(self._score_edges, 
                                               [(vertex, neighbors, N, scored_graph) for vertex in range(N)])
        else:
            graphs = [self._score_edges(vertex, neighbors, N, scored_graph) for vertex in range(N)]
            
        final_graph = normalize(np.sum(graphs, axis=0), norm='l1', axis=1, copy=False)
        return final_graph
        
    
    def _score_edges(self, vertex, neighbors, N, scored_graph):
        C = lil_matrix((N,N))
        for v in range(N):
            for w in range(v+1,N):
                if (v in neighbors[vertex]) and (w in neighbors[vertex]):
                    edge_prob = scored_graph[v,w]
                    if edge_prob >= self.epsilon:
                        C[w,v] += edge_prob
                        C[v,w] += edge_prob
        
        return csr_matrix(C)
    
    
    
class modularity_knn_graph:
    def __init__(self, metric='euclidean', graph_type='symmetric', 
                 clustering_alg ='louvain', clustering_iterations=5):
        
        self.metric = metric
        self.best_modularity =-1
        self.best_network = nx.Graph()
        self.best_part =None
        self.best_k =2
        self.graph_type = graph_type
        self.clustering_alg = clustering_alg
        self.clustering_iterations = clustering_iterations
         
        
    def learn_graph(self, data_matrix):
    
        for n in range(1, np.int(np.floor(np.log2(data_matrix.shape[0])))):
            k = 2**n
            kNN = kneighbors_graph(data_matrix, metric=self.metric, mode='connectivity', n_neighbors=int(k))
            if self.graph_type == 'symmetric':
                kNN = kNN.minimum(kNN.T)
            elif self.graph_type=='assymetric':
                kNN = kNN.maximum(kNN.T)
#            kNN = dist_to_sim().convert(kNN)
            network = nx.from_scipy_sparse_matrix(kNN)
            if self.clustering_alg =='leiden':                
                curr_part, avg_modularity = self._get_leiden_partition_and_modularity(kNN)
                random_kNN = nx.to_scipy_sparse_matrix(self._randomize_network(network, self.graph_type))
                curr_random_part, avg_random_modularity = self._get_leiden_partition_and_modularity(random_kNN)
                curr_modularity = avg_modularity - avg_random_modularity
                
            elif self.clustering_alg =='louvain':
                curr_part, avg_modularity = self._get_louvain_partition_and_modularity(kNN)
                random_kNN = nx.to_scipy_sparse_matrix(self._randomize_network(network, self.graph_type))
                curr_random_part, avg_random_modularity = self._get_louvain_partition_and_modularity(random_kNN)
                curr_modularity = avg_modularity - avg_random_modularity
            else:
               raise Exception('select valid clustering algorithm. Options are leiden and louvain')
             
            #network = nx.from_scipy_sparse_matrix(kNN)
            #curr_part = community.best_partition(network)
            #random_network = self._randomize_network(network)
            #curr_random_part = community.best_partition(random_network)
            #curr_modularity = community.modularity(curr_part, network) - community.modularity(curr_random_part, random_network)
#            currModularity = community.modularity(currPart, network) - self._random_modularity(network)
            if curr_modularity > self.best_modularity:
                self.best_modularity = curr_modularity
                self.best_network = network
                self.best_part = curr_part
                self.best_k = k

        return nx.to_scipy_sparse_matrix(self.best_network)
    
    
    def _random_modularity(self, network):
        S = network.number_of_nodes()
        p = nx.density(network)
        return (1-2/np.sqrt(S))*(2/(p*S))**(2/3)
    
    
    def _get_louvain_partition_and_modularity(self, kNN):
        cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=self.clustering_iterations)]
        parts = [clstr.fit(kNN).labels_ for clstr in cluster_iterations]
        modularities = [modularity(kNN, part) for part in parts]
        best_part = parts[np.argmax(modularities)]
        avg_modularity = np.mean(modularities)
        return best_part, avg_modularity
        
    def _get_leiden_partition_and_modularity(self, kNN):
        import leidenalg as la
        graphs =[self._scipy_to_igraph(kNN)]*self.clustering_iterations
        leiden_results = [la.find_partition(G, la.RBConfigurationVertexPartition, 
                                            resolution_parameter=1.0, n_iterations = -1) 
                          for G in graphs]
        modularities =[partition.modularity for partition in leiden_results]
        parts = [np.array(partition.membership) for partition in leiden_results]
        best_part = parts[np.argmax(modularities)]
        avg_modularity = np.mean(modularities)
        return best_part, avg_modularity
    
    def _scipy_to_igraph(self, matrix):
        from igraph import Graph
        if self.graph_type =='directed':
            directed=True
        else:
            directed = False
        sources, targets = matrix.nonzero()
        weights = matrix[sources, targets]
        return Graph(list(zip(sources, targets)), directed=directed, edge_attrs={'weight': weights})
    
    def _randomize_network(self, network, graph_type):
        n = network.number_of_nodes()
        p = nx.density(network)
        if graph_type =='directed':
            return nx.fast_gnp_random_graph(n, p, directed=True)
        else:
            return nx.fast_gnp_random_graph(n, p, directed=False)


    
class l1_sparse_graph:
    def __init__(self, lmbda=0.001):
        self.lmbda = lmbda
        
        
    def learn_graph(self, data_matrix):
        pass
    


class l0_low_rank_graph:
    def __init__(self, lmbda=0.001):
        self.lmbda = lmbda
        
        
    def learn_graph(self, data_matrix):
        pass



class tensor_graph_diffusion:
    def __init__(self, T=20, alpha=0.9, nearest_neighbors=10, tau=0, 
                 kernel_graph=None):
        
        self.T = T
        self.alpha =0.9
        self.k = nearest_neighbors
        self.tau = tau
        self.kernel_graph = kernel_graph
        
        
    def learn_graph(self, full_graph):
        
        if issparse(full_graph):
            Q = normalize(full_graph, norm='l1', axis=1, copy=False).todense()
        else:
            Q = normalize(full_graph, norm='l1', axis=1, copy=False)
        
        N = Q.shape[0]
        if self.kernel_graph is None:
            idxs = np.flip(np.argsort(Q, axis=1), axis=1)[:,1:self.k+1]
            graph = np.zeros((N,N))
            graph[np.arange(N)[:,None], idxs] = Q[np.arange(N)[:,None], idxs]
            graph = normalize(graph, norm='l1', axis=1, copy=False)
            self.kernel_graph = csr_matrix(graph)
            Q = csr_matrix(Q)
        else:
            self.kernel_graph = csr_matrix(normalize(self.kernel_graph, norm='l1', axis=1, copy=False))
            Q = csr_matrix(Q)

        for t in range(self.T+1):
            Q = self.alpha*(self.kernel_graph * Q) * self.kernel_graph.transpose() + (1-self.alpha)*eye(N)
            
        Q.setdiag(0)
        
        if self.tau == 'mean':
            self.tau = np.mean(Q.data)
        Q.data[Q.data < self.tau] = 0
        Q.eliminate_zeros()
        return Q 



class network_enhancement:
    def __init__(self, T=20, alpha=0.9, nearest_neighbors=10, tau=0, graph_type='symmetric',
                 kernel_graph=None):
        
        self.T = T
        self.alpha =0.9
        self.k = nearest_neighbors
        self.tau = tau
        self.graph_type = graph_type
        self.kernel_graph = kernel_graph
        
    
    def learn_graph(self, full_graph):
        
        if self.graph_type =='symmetric':
            Q = full_graph.minimum(full_graph.T)
        elif self.graph_type =='assymetric':
            Q = full_graph.maximum(full_graph.T)
        else:
            raise Exception('{} is not a support graph type, must be symmetric or assymetric'.format(self.graph_type))
        
        D = diags(np.array(np.sqrt(np.sum(Q, axis=1))).ravel())
        Q = self._DSM(Q)
        
        N = Q.shape[0]
        if self.kernel_graph is None:
            full_graph = Q.todense()
            idxs = np.flip(np.argsort(full_graph, axis=1), axis=1)[:,1:self.k+1]
            graph = np.zeros((N,N))
            graph[np.arange(N)[:,None], idxs] = full_graph[np.arange(N)[:,None], idxs]
            self.kernel_graph= self._DSM(csr_matrix(graph))
        else:
            self.kernel_graph = self._DSM(self.kernel_graph)
            
        for t in range(self.T+1):
            Q = self.alpha*(self.kernel_graph * Q * self.kernel_graph) + (1-self.alpha)*self.kernel_graph
            
        Q.setdiag(0)
        Q = D*Q*D
        
        if self.tau == 'mean':
            self.tau = np.mean(Q.data)
        Q.data[Q.data < self.tau] = 0
        Q.eliminate_zeros()
        return Q      
    
    
    def _DSM(self, graph):
        graph = graph.todense()
        graph = normalize(graph, norm='l1', axis=1, copy=False)
        col_sums= np.sum(graph, axis=0)
        temp_graph = np.divide(graph, col_sums**0.5, out=np.zeros_like(graph), where=col_sums!=0)
        return csr_matrix(temp_graph @ temp_graph.T)
 
'''
Multi-modal Graph Leanring Techniques
'''
    
    
class edge_score_graph:
    def __init__(self, scoring='bray_curtis', epsilon='mean', multi_process = False):
        
        self.scoring = scoring
        self.epsilon = epsilon
        self.multi_process = multi_process
            
        
    def learn_graph(self, input_graphs):
        self.input_graphs = input_graphs
        
        if self.multi_process:
            with mp.Pool(processes=len(input_graphs)) as pool:
                scored_graphs = pool.map(self._score_graph, 
                                               [graph for graph in self.input_graphs])
        else:
            scored_graphs =[self._score_graph(graph) for graph in self.input_graphs]
            
        combined_graph = np.sum(scored_graphs)/len(scored_graphs)
        if self.epsilon == 'mean':
            self.epsilon = np.mean(combined_graph.data)
        combined_graph.data[combined_graph.data < self.epsilon] = 0
        combined_graph.eliminate_zeros()

        final_graph = normalize(combined_graph, norm='l1', axis=1, copy=False)
        return final_graph
    
    
    def _score_graph(self, graph):
        N = graph.shape[0]
        neighbors = {vertex:graph.getrow(vertex).nonzero()[1] for vertex in range(N)}
        vectors = {vertex:graph[vertex] for vertex in range(N)}
        return np.sum([self._score_edges(vertex, neighbors, vectors, N) for vertex in range(N)])
        
    
    def _score_edges(self, vertex, neighbors, vectors, N):
        scorer = scoring()
        scored_graph = scorer.score(vertex, neighbors, vectors, N, self.scoring)
        return scored_graph
 
  
    
class annealed_graph:
    def __init__(self, scoring='bray_curtis', multi_process = False):
        
        self.scoring = scoring
        self.multi_process = multi_process
           
        
    def learn_graph(self, input_graphs):
        self.input_graphs = input_graphs
            
        if self.multi_process:
            with mp.Pool(processes=len(input_graphs)) as pool:
                scored_graphs = pool.map(self._score_graph, 
                                               [graph for graph in self.input_graphs])
        else:
            scored_graphs = [self._score_graph(graph) for graph in self.input_graphs]
            
#        smooth = smooth_neighborhood_graph()
#        return smooth.learn_graph(scored_graphs)
            
#        intersect_graph = scored_graphs[0]
#        for graph in scored_graphs[1:]:
#            intersect_graph = intersect_graph.minimum(graph)    
#        return intersect_graph
        
#        mask = scored_graphs[0] > 0
#        for graph in scored_graphs[1:]:
#            mask = mask.multiply(graph > 0) 
#        intersect_graph = np.sum(scored_graphs).multiply(mask) / len(scored_graphs)
#        union_graph = np.sum(scored_graphs) / len(scored_graphs)
            
#        A = normalize(intersect_graph.todense(), norm='l1', axis=1, copy=False)
#        N = A.shape[0]
#        Q = A.copy()
#        for t in range(self.T):
#            Q = np.matmul(np.matmul(A, Q), np.transpose(A)) + np.eye(N)
            
#        np.fill_diagonal(A,0)
#        return csr_matrix(normalize(A, norm='l1', axis=1, copy=False))
#            
#        combined_graph = np.sum(scored_graphs)
#        combined_graph = combined_graph.todense()
#        mask = mask.todense()
#        final_graph = np.multiply(combined_graph, mask)
#        N = final_graph.shape[0]
#        density = np.count_nonzero(final_graph)/(N*(N-1))
#        for graph in scored_graphs:
#            graph = graph.todense()
#            auxilliary_edges = np.multiply(graph, ~mask)
#            inv_score = 1-auxilliary_edges
#            auxilliary_edges[auxilliary_edges<inv_score] = 0
#            boltzmann_factor = np.exp(-1*(1-(auxilliary_edges-inv_score))/(1-density)) >= np.random.sample((N,N))
#            auxilliary_edges = np.multiply(auxilliary_edges, boltzmann_factor)
#            final_graph = final_graph + auxilliary_edges
#        return csr_matrix(final_graph/ len(scored_graphs))
        
        combined_graph = np.sum(scored_graphs)/len(scored_graphs)
        N = combined_graph.shape[0]
        density = np.mean([len(graph.data)/(N*(N-1)) for graph in scored_graphs])
#        density = len(combined_graph.data)/(N*(N-1)) 
        for i in range(len(combined_graph.data)):
            score = combined_graph.data[i]
            inv_score = 1- score
            if score < inv_score:
                if np.exp(-1*(inv_score-score)/(1-density)) < np.random.sample():
                    combined_graph.data[i] =0
            
        combined_graph.eliminate_zeros()
        final_graph = normalize(combined_graph, norm='l1', axis=1, copy=False)
        return final_graph
    

    def _score_graph(self, graph):
        '''
        Still need to create a fast means of scoring on sparse matrices
        '''
        from scipy.spatial.distance import pdist, squareform
        new_graph = squareform(pdist(graph.todense(), metric='braycurtis'))
        new_graph = 1-new_graph
        new_graph = np.nan_to_num(new_graph)
        np.fill_diagonal(new_graph, 0)
        return csr_matrix(new_graph)
#        N = graph.shape[0]
#        neighbors = {vertex:graph.getrow(vertex).nonzero()[1] for vertex in range(N)}
#        vectors = {vertex:graph[vertex] for vertex in range(N)}
#        density = len(graph.data)/(N*(N-1))
#        return np.sum([self._score_edges(vertex, neighbors, vectors, N, density) for vertex in range(N)])
    
    
    def _score_edges(self, vertex, neighbors, vectors, N, density):
        scorer = scoring()
        scored_graph = scorer.score(vertex, neighbors, vectors, N, self.scoring)
        return scored_graph
#        scorer = scoring()
#        scored_graph = scorer.score(vertex, neighbors, vectors, N, self.scoring)
#        for i in range(len(scored_graph.data)):
#            score = scored_graph.data[i]
#            inv_score = 1- score
#            if score < inv_score:
#                if np.exp(-1*(inv_score-score)/(1-density)) < np.random.sample():
#                    scored_graph.data[i] =0
#            if score < inv_score:
#                scored_graph.data[i]=0
#            elif np.exp(-1*(1-(score-inv_score))/(1-density)) < np.random.sample():
#                scored_graph.data[i] =0
#        return scored_graph
    

    
class smooth_neighborhood_graph:
    def __init__(self, gamma = 2, mu = 0.8, tau=0, 
                 multi_process = False):
        
        self.multi_process = multi_process
        self.gamma = gamma
        self.mu = mu
        self.tau = tau
            
        
    def learn_graph(self, input_graphs):
        self.input_graphs = input_graphs
        N = self.input_graphs[0].shape[0]
        
        self.input_graphs = [graph.todense() for graph in self.input_graphs]
        
        alpha = np.ones(len(self.input_graphs)) * (1/len(self.input_graphs))
        converged = False
        J_prev=0    
        L= []
        for idx in range(len(self.input_graphs)):
            A = self.input_graphs[idx]
            D= np.multiply(np.sum(A,1), np.eye(N))
            L.append(D - A)
            
        while not converged:
            Y = 0
            for idx in range(len(self.input_graphs)):
                Y += alpha[idx]**self.gamma *L[idx] + self.mu * np.eye(N)
            Y = self.mu * np.linalg.inv(Y)
            
            for idx in range(len(self.input_graphs)):
                quad = np.matmul(np.matmul(np.transpose(Y), L[idx]), Y)
                alpha[idx] = (np.trace(quad))**(1/(1-self.gamma))
                normalizer = np.sum(alpha)
                alpha /= normalizer
                
            J = self.mu * np.linalg.norm( Y- np.eye(N), ord='fro')**2
            for idx in range(len(self.input_graphs)):
                quad = np.matmul(np.transpose(Y), np.matmul(L[idx], Y))
                J += alpha[idx]**self.gamma * np.trace(quad)
                
            if np.abs(J_prev-J) < 1e-6:
                converged = True
            else:
                J_prev= J
                
        np.fill_diagonal(Y,0)
        final_graph = csr_matrix(normalize(Y, norm='l1', axis=1, copy=False))
        if self.tau == 'mean':
            self.tau = np.mean(final_graph.data)
        final_graph.data[final_graph.data < self.tau] = 0
        final_graph.eliminate_zeros()
        return final_graph
    
    
    
class cross_diffusion_graph:
    def __init__(self, T=10, multi_process = False, tau=0, alpha=0.9, iter_thresh=0,
                 nearest_neighbors=10, kernel_graphs=None):
        
        self.tau= tau
        self.iter_thresh = iter_thresh
        self.multi_process = multi_process
        self.T = T
        self.k = nearest_neighbors
        self.alpha = alpha
        self.kernel_graphs = kernel_graphs
        
        
    def learn_graph(self, full_graphs):
        
        if self.kernel_graphs is None:
            self.kernel_graphs =[]
            for graph in full_graphs:
                graph = graph.todense()
                N = graph.shape[0]
                idxs = np.flip(np.argsort(graph, axis=1), axis=1)[:,1:self.k+1]
                kernel_graph = np.zeros((N,N))
                kernel_graph[np.arange(N)[:,None], idxs] = graph[np.arange(N)[:,None], idxs]
                self.kernel_graphs.append(csr_matrix(normalize(kernel_graph, norm='l1', axis=1, copy=False)))
        else:
            graphs =[]
            for graph in self.kernel_graphs:
                graphs.append(normalize(graph, norm='l1', axis=1, copy=True))
            self.kernel_graphs = graphs
        
        self.P = [normalize(graph, norm='l1', axis=1, copy=True) for graph in full_graphs]
        for iteration in range(self.T):
            P_new = []
            for v in range(len(self.P)):
                S_v = self.kernel_graphs[v]
                P_k = np.sum([self.P[idx] for idx in range(len(self.P)) if idx != v], axis=0)/(len(self.P)-1)
                result = self.alpha * (S_v @ P_k @ S_v.T) + (1-self.alpha) * eye(P_k.shape[0])
                # result = self.alpha * (S_v @ P_k @ S_v.T) + (1-self.alpha) * full_graphs[v]
                # threshold = np.quantile(result.data, self.iter_thresh)
                # result.data[result.data < threshold] = 0
                # result.eliminate_zeros()
                P_new.append(normalize(result, norm='l1', axis=1, copy=True))
                
            self.P = P_new
        
        P_final = np.sum(np.array(self.P), axis=0)/ len(self.P)
        P_final.setdiag(0)
        final_graph = normalize(P_final, norm='l1', axis=1, copy=False)
        if self.tau == 'mean':
            self.tau = np.mean(final_graph.data)
        final_graph.data[final_graph.data < self.tau] = 0
        final_graph.eliminate_zeros()
        return final_graph
    

        
class cross_enhancement_graph:
    def __init__(self, T=10, multi_process = False, tau=0, alpha=0.9,
                 nearest_neighbors=10, kernel_graphs=None):
        
        self.tau= tau
        self.alpha = alpha
        self.multi_process = multi_process
        self.T = T
        self.k = nearest_neighbors
        self.kernel_graphs = kernel_graphs
        self._kernel_graphs_dsm = False
        
        
    def learn_graph(self, full_graphs):
        N= full_graphs[0].shape[0]
        self.degree_matrices = [diags(np.array(np.sqrt(np.sum(graph, axis=1))).ravel()) for graph in full_graphs]
        self.full_graphs = [self._DSM(graph) for graph in full_graphs]
        
        if self.kernel_graphs is None:
            self.kernel_graphs =[]
            for graph in self.full_graphs:
                graph = graph.todense()
                N = graph.shape[0]
                idxs = np.flip(np.argsort(graph, axis=1), axis=1)[:,1:self.k+1]
                raw_graph = np.zeros((N,N))
                raw_graph[np.arange(N)[:,None], idxs] = graph[np.arange(N)[:,None], idxs]
                raw_graph = csr_matrix(raw_graph)
                self.kernel_graphs.append(self._DSM(raw_graph))

        elif self._kernel_graphs_dsm == False:
            graphs =[]
            for raw_graph in self.kernel_graphs:
                graphs.append(self._DSM(raw_graph))
            
            self.kernel_graphs = graphs
            self._kernel_graphs_dsm = True
        
        self.P = [graph.copy() for graph in self.full_graphs]
        for iteration in range(self.T):
            P_new = []
            for v in range(len(self.P)):
                S_v = self.kernel_graphs[v]
                P_k = np.sum(np.array([self.P[idx] for idx in range(len(self.P)) if idx != v]), axis=0)/(len(self.P)-1)
                result = self.alpha * (S_v @ P_k @ S_v) + (1-self.alpha)*S_v
                P_new.append(result)
                
            self.P = P_new
        
        self.P = [self.degree_matrices[i] @ self.P[i] @ self.degree_matrices[i] for i in range(len(self.P))]
        final_graph = np.sum(np.array(self.P), axis=0)/ len(self.P)
        final_graph.setdiag(0)
        
        if self.tau == 'mean':
            self.tau = np.mean(final_graph.data)
        final_graph.data[final_graph.data < self.tau] = 0
        final_graph.eliminate_zeros()
        return final_graph
    
    def _DSM(self, graph):
        graph = graph.todense()
        graph = normalize(graph, norm='l1', axis=1, copy=False)
        col_sums= np.sum(graph, axis=0)
        temp_graph = np.divide(graph, col_sums**0.5, out=np.zeros_like(graph), where=col_sums!=0)
        return csr_matrix(temp_graph @ temp_graph.T)


    
class annealed_sparsified_graph:
    '''
    Note: this will need its own way of scoring edges, since it does not
    need scores for potential edges
    '''
    def __init__(self, max_iter=10, scoring='bray_curtis', max_stall=5,
                 tol = 0.01, sample_proportion=0.3,  multi_process = False, n_processes=4):
        
        self.scoring= scoring
        self.max_iter = max_iter
        self.max_stall = max_stall
        self.tol = tol
        self.sample_proportion = sample_proportion
        self.multi_process = multi_process
        self.n_processes=n_processes
        
        
    def learn_graph(self, original_graph):        
        N= original_graph.shape[0]
        current_graph = lil_matrix(original_graph)
        curr_stall=0
        curr_iter=1
        test_density =[]
        sample_size = int(np.floor(self.sample_proportion *N))
        
        while (curr_iter < self.max_iter) and (curr_stall < self.max_stall):
            C = current_graph.copy()
            
            density = current_graph.count_nonzero()/(N*(N-1))
            test_density.append(density)
            
            neighbors = {vertex:C.getrow(vertex).nonzero()[1] for vertex in range(N)}
            vectors = {vertex:C[vertex] for vertex in range(N)}
                
            if self.multi_process:
                with mp.Pool(processes=self.n_processes) as pool:
                    graphs = pool.starmap(self._score_edges, 
                                                   [(vertex, neighbors, vectors, N) for vertex in range(N)])
            else:
                graphs = [self._score_edges(vertex, neighbors, vectors, N) for vertex in range(N)]
            
            scored_graph = np.sum(graphs)
            scored_graph = scored_graph.multiply(current_graph >0)
            
            for u in np.random.choice(range(N), size=sample_size, replace=False):
                N_u = neighbors[u]
                scores = scored_graph[u].data
                idxs = np.argsort(scores)
                for idx in idxs[:-1]:
                    v = N_u[idx]
                    score = scores[idx]
                    inv_score= 1-score
                    if inv_score > score:
                        C[u,v] = 0
                    elif np.exp(-1*(score-inv_score)/(density)) > np.random.sample():
                        C[u,v] = 0
                    else:
                        break
    
            new_density = C.count_nonzero()/(N*(N-1))
            if np.abs(density-new_density) < self.tol:
                curr_stall +=1
            else:
                current_graph = C
                curr_stall = 0
                
            curr_iter += 1
                
        final_graph = csr_matrix(current_graph) 
        return (final_graph, test_density)
    
    
    def _score_edges(self, vertex, neighbors, vectors, N):
        scorer = scoring()
        return scorer.score(vertex, neighbors, vectors, N, self.scoring)
  
'''
Axuiliary Classes
'''  
    
class thermodynamic_measures:     
    def approx_entropy(G):
        G= G.toarray()
        try:
            check_symmetric(G, raise_exception=True)
            directed= False
        except ValueError:
            directed = True
        N = G.shape[0]
        if directed:
            out_degree = np.count_nonzero(G,axis=1)
            in_degree = np.count_nonzero(G, axis=0)
            ratio_match = np.nan_to_num(np.multiply(G>0, in_degree.T / np.outer(in_degree, out_degree**2)).astype(float), nan=0.0, posinf=0.0)
            out_degree_match = np.multiply((G *G.T)>0, np.outer(out_degree, out_degree)).astype(float)
            inversion_term = np.ones_like(ratio_match)
            ratio_term = np.divide(inversion_term, ratio_match, 
                                    out=np.zeros_like(inversion_term), where=ratio_match!=0)
            out_degree_term = np.divide(inversion_term, out_degree_match, 
                                    out=np.zeros_like(inversion_term), where=out_degree_match!=0)
            return 1- 1/N - (1/(2*N**2))*(np.sum(ratio_term)+np.sum(out_degree_term))
        else:
            degree = np.count_nonzero(G,axis=1)
            degree_match = np.multiply(G>0, np.outer(degree, degree)).astype(float)/2
            inversion_term = np.ones_like(degree_match)
            degree_term = np.divide(inversion_term, degree_match, 
                                    out=np.zeros_like(inversion_term), where=degree_match!=0)
            return 1- 1/N - (1/(N**2))*np.sum(degree_term)
    
    def exact_entropy(G):
        G= G.toarray()
        N= G.shape[0]
        D= np.diag(np.ravel(np.sum(G, axis=1)))
        D_inv = np.nan_to_num(np.diag(np.ravel(np.sum(G, axis=1))**-0.5))
        L = np.nan_to_num(D_inv @ (D-G) @ D_inv)
        eigenvalues = np.real(np.linalg.eig(L)[0])
        eigenvalues[eigenvalues<1e-9] =0 
        eigenvalues = eigenvalues[eigenvalues.nonzero()]
        return -1* np.sum(np.nan_to_num(np.multiply(eigenvalues/N, np.log(eigenvalues/N))))
    
    def internal_energy(G):
        G= np.array(G)
        N = G.shape[0]
        degree = np.count_nonzero(G, axis=1)
        return np.sum(np.multiply(N, degree))
    
    def temperature(G_1,G_2):
        G_1= np.array(G_1)
        G_2= np.array(G_2)
        T_inv = (thermodynamic_measures.approx_entropy(G_2) - thermodynamic_measures.approx_entropy(G_1))/ \
        (thermodynamic_measures.internal_energy(G_2) - thermodynamic_measures.internal_energy(G_1))
        
        return 1/T_inv
    
 
           
class scoring:
    def score(self, vertex, neighbors, vectors, N, format):
        scorer = self._get_scorer(format)
        return scorer(vertex, neighbors, vectors, N)
    
    
    def _get_scorer(self, format):
        if format == 'weighted_jaccard':
            return self._weighted_jaccard
        elif format == 'weighted_adamic_adar':
            return self._weighted_adamic_adar
        elif format == 'weighted_resource_allocation':
            return self._weighted_resource_allocation
        elif format == 'bray_curtis':
            return self._bray_curtis
        else:
            raise ValueError(format)
    
    
    def _weighted_jaccard(self, u, neighbors, vectors, N):
        scored_graph = lil_matrix((N,N))
        N_u = vectors[u]
        for v in range(u+1, N):
            N_v = vectors[v]
            scored_graph[u,v] = N_v.minimum(N_u).sum()/N_v.maximum(N_u).sum()
            scored_graph[v,u] += scored_graph[u,v]
        
        scored_graph =csr_matrix(scored_graph)
        scored_graph.data[scored_graph.data < 0.0000001] =0
        scored_graph.eliminate_zeros()
        return scored_graph
      
        
    def _weighted_adamic_adar(self, u, neighbors, vectors, N):
        scored_graph = lil_matrix((N,N))
        N_u = vectors[u]
        for v in range(u+1, N):
            N_v = vectors[v]
            for z in neighbors[v]:
                if (z != u) and (z in neighbors[u]):
                    scored_graph[u,v] += (N_u[0,z] + N_v[0,z])/(np.log(1+vectors[z].sum()))
                    scored_graph[v,u] += scored_graph[u,v]
        
        scored_graph =csr_matrix(scored_graph)
        scored_graph.data[scored_graph.data < 0.0000001] =0
        scored_graph.eliminate_zeros()
        scored_graph.data = np.exp(-1/scored_graph.data)
        return scored_graph
    
    
    def _weighted_resource_allocation(self, u, neighbors, vectors, N):
        scored_graph = lil_matrix((N,N))
        N_u = vectors[u]
        for v in range(u+1, N):
            N_v = vectors[v]
            for z in neighbors[v]:
                if (z != u) and (z in neighbors[u]):
                    scored_graph[u,v] += (N_u[0,z] + N_v[0,z])/(vectors[z].sum())
                    scored_graph[v,u] += scored_graph[u,v]
        
        scored_graph =csr_matrix(scored_graph)
        scored_graph.data[scored_graph.data < 0.0000001] =0
        scored_graph.eliminate_zeros()
        scored_graph.data = np.exp(-1/scored_graph.data)
        return scored_graph
    
    
    def _bray_curtis(self, u, neighbors, vectors, N):
        scored_graph = lil_matrix((N,N))
        N_u = vectors[u]
        N_u_sum = N_u.sum()
        for v in range(u+1, N):
            N_v = vectors[v]
            denominator = N_u_sum +N_v.sum()
            if denominator > 0:
                scored_graph[u,v] = 1-(abs(N_v-N_u).sum())/(denominator)
                scored_graph[v,u] = scored_graph[u,v]
          
        scored_graph =csr_matrix(scored_graph)
        scored_graph.data[scored_graph.data < 0.0000001] =0
        scored_graph.eliminate_zeros()
        return scored_graph



class dist_to_sim:
    def convert(self, graph, conversion='guassian', mu=0.5):
        self.mu = mu
        converter = self._get_converter(conversion)
        return converter(graph)
    
    def _get_converter(self, conversion):
        if conversion == 'guassian':
            return self._guassian
        elif conversion == 'one_minus':
            return self._one_minus
        else:
            raise ValueError(conversion+" not found")
        
    def _one_minus(self, graph):
        graph.data = 1-graph.data
        return graph
    
    def _guassian(self, graph):
        '''
        Convert using heat/ RBF kernel with local estimates of sigma by standard
        deviation of outgoung edges for each vertex
        '''
        affinity_graph = graph.copy().tolil()
        distance_graph = graph.copy().tocoo()
        means = np.mean(graph, axis=1)
        for i,j,v in zip(distance_graph.row, distance_graph.col, distance_graph.data):
            affinity_graph[i,j] = np.exp((-1*v**2)/(self.mu*((means[i]+means[j]+v)/3)))
            
        return affinity_graph.tocsr()
       
    
#class sample_graph:
#    """
#    STILL UNDER DEVELOPMENT. idea was to use the same idea as hard disks in a
#    box from early boltzmann research
#    """
#    def __init__(self, distance_matrix, k_initial=0.5, samples=100000):
#        self.X = np.array(distance_matrix)
#        self.samples=samples
#        N = self.X.shape[0]
#        self.K = np.ones(N) * np.floor(k_initial*N)
#            
#    def learn_graph(self):
#        A= self.X.copy()
#        N= A.shape[0]
#        idxs = []
#        for i in range(N):
#            nonzero_entries = np.nonzero(A[i,:])[0]
#            sorted_nonzero_idxs = np.argsort(A[i,nonzero_entries])
#            idxs.append(nonzero_entries[sorted_nonzero_idxs]) 
#          
#        neighbors =[self.K]
#        for sample in range(self.samples):
#            node = np.random.choice(range(N))
#            if self.K[node] == 1:
#                delta = 1
#            elif self.K[node] == N-1:
#                delta = -1
#            else:
#                delta = np.random.choice([-1,1])
#            k_candidate = self.K.copy()
#            k_candidate[node] += delta
#            
##            if delta < 0:
##                old_score = self._compute_prob(self.K, idxs, A, node)
##                new_score= 1-old_score
##            else:
##                new_score = self._compute_prob(k_candidate, idxs, A, node)
##                old_score = 1-new_score
##            if new_score > old_score:
##                self.K = k_candidate
##                neighbors.append(self.K)
##            elif np.exp(-1*(1-new_score/(1-np.sum(self.K)/(N*(N-1))))) > np.random.sample():
##                self.K = k_candidate
##                neighbors.append(self.K)
#                
##            if delta < 0:
##                score = self._compute_prob(self.K, idxs, A, node)
##            else:
##                score = self._compute_prob(k_candidate, idxs, A, node)
##                
##            if np.exp(-1*(1-score/(1-np.sum(self.K)/(N*(N-1))))) > np.random.sample():
##                self.K = k_candidate
##                neighbors.append(self.K)
#            if delta >0:
#                score = self._compute_prob(k_candidate, idxs, A, node)
#                inv_score = 1- score
#            else:
#                inv_score = self._compute_prob(self.K, idxs, A, node)
#                score = 1- inv_score
#            
#            if score > inv_score:
#                self.K = k_candidate
#                neighbors.append(self.K)
#            elif np.exp(-1*(inv_score-score)/(
#                    min(1-np.sum(self.K)/(N*(N-1)), np.sum(self.K)/(N*(N-1))
#                    ))) > np.random.sample():
#                self.K = k_candidate
#                neighbors.append(self.K)
#                
#        self.K = np.floor((np.mean(neighbors, axis=0)))
##        self.K = neighbors[-1]
#        C = self._compute_graph(idxs, A)
#        degree = np.sum(C, axis=1)
#        D= degree[:, np.newaxis] 
#        return np.divide(C, D, out=np.zeros_like(C), where=D!=0) 
#                
#    def _compute_prob(self, K, idxs, A, u):
#        N= A.shape[0]
#        k_u = int(K[u])
#        v = idxs[u][k_u-1]
#        N_u = np.zeros(N)
#        N_u[idxs[u][0:k_u-1]] = A[u,idxs[u][0:k_u-1]]
#        k_v = int(K[v])
#        N_v = np.zeros(N)
#        N_v[idxs[v][0:k_v-1]] = A[v,idxs[v][0:k_v-1]]
##        numerator = (A[u,v]+np.sum(np.minimum(N_u,N_v)))/(
##                    A[u,v]+np.sum(np.maximum(N_u,N_v)))
##        denominator = (N-k_u)/N
##        return np.exp(-1*(numerator/denominator))
#        return (A[u,v]+np.sum(np.minimum(N_u,N_v)))/(
#                    A[u,v]+np.sum(np.maximum(N_u,N_v)))
#        
#    def _compute_graph(self, idxs, A):
#        N= A.shape[0]
#        C = np.zeros((N,N))
#        for u in range(N):
#            N_u = np.zeros(N)
#            k_u = int(self.K[u])
#            N_u[idxs[u][0:k_u-1]] = A[u,idxs[u][0:k_u-1]]
#            for v in N_u.nonzero()[0]:
#                if u != v:
#                    N_v = np.zeros(N)
#                    k_v = int(self.K[v])
#                    N_v[idxs[v][0:k_v-1]] = A[v,idxs[v][0:k_v-1]]
#                    C[u,v] += (A[u,v]+np.sum(np.minimum(N_u,N_v)))/(
#                            A[u,v]+np.sum(np.maximum(N_u,N_v)))
#                    
#        return C
        
        
